# Copyright (C) 2001 by Dr. Dieter Maurer <dieter@handshake.de>
# D-66386 St. Ingbert, Eichendorffstr. 23, Germany
#
#			All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice and this permission
# notice appear in all copies, modified copies and in
# supporting documentation.
# 
# Dieter Maurer DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL Dieter Maurer
# BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
# DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
# PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
'''Wizzard

 A Wizzard supports the step-by-step execution of a complex
 task through a series of forms. Each form is represented
 by a dispatcher which maintains the forms state.

 The Wizzard is able to set and read the state of all controlled
 forms. To set the state, it supports the methods 'new', 'edit', 'clear'
 and 'update'.
 'getValues' returns the complete state in a single mapping object.
 If there are name conflicts between various forms, forms with
 a lower index take preceedance.
 'getCollection' is a synonym to avoid conflicts with the
 'getValues' method of dispatchers.

 The tokens property 'Dispatchers__' holds the list of controlled
 dispatchers.

 Wizzards derive from folders. Usually, the dispatchers are the Wizzards
 children. However, they can also be acquired. But, handling is more
 difficult then.


 Properties

   'Dispatchers__'  --  ('tokens') list of dispatcher ids
   
     Each id determines a dispatcher controlled by the Wizzard.

     Ids are resolved by 'restrictedTraverse'.
     This means, they can be URL fragments.


 Methods

   'new' -- usually used to collect attributes for a new object

     It calls each dispatchers 'new' method and
     then renders a dispatcher.

   'edit' -- usually used to change the attributes of an existing object

     It calls each dispatchers 'edit' method and
     then renders a dispatcher.

   'clear' -- used to clear all remembered values (with the exception of fixed
     variables)

     It calls each dispatchers 'clear' method and
     then renders a dispatcher.

   'update' -- update remembered values

     It calls each dispatchers 'update' and then renders a dispatcher

   'getValues' -- the remembered variables as a mapping object

   'getCollection' -- a synonym for 'getValues'


 Arguments

    The methods 'new', 'edit', 'clear' and 'update' all take
    DTML like arguments.

    The argument 'WizzardDestination__' controls which dispatcher
    is rendered, see **Rendering**.

    The argument 'WizzardURL__' gives the Wizzards URL provided
    it can not be determined from the context. See **Rendering**
    for details.


 Rendering

    At the end of execution of 'new', 'edit', 'clear' and 'update',
    the Wizzard usually renders a dispatcher. The argument
    'WizzardDestination__' controls which dispatcher is rendered.

    If there is not such argument, then the first dispatcher
    is rendered. If the argument is empty, then rendering is
    suppressed. Otherwise, it must be the id of one of the
    controlled dispatchers. And this dispatcher is rendered.

    Rendering is performed by calling the dispatchers 'show' method.

    If the rendered object uses relative URLs, then it is
    important that the base URL is correct for their resolution.
    Without further action, this would be almost always wrong for
    objects rendered via a Wizzard. Therefore, the base
    URL must be explicitly set.
    If the methods are directly called via HTTP, then the
    Wizzard can derive its URL from the request context ('URL1').
    However, when called indirectly (e.g. via DTML), then
    this may no longer work.
    In these cases, the argument 'WizzardURL__' must be used
    to inform the Wizzard about its URL.
    *Note* that we can not use 'absolute_url', because it discards
    all context from the URL.
    '''



from OFS.Folder import Folder
from AccessControl.SecurityInfo import ClassSecurityInfo
from Globals import InitializeClass, DTMLFile
from Acquisition import Explicit


from Products.Dispatcher.HelpSupport import HelpSTXTopic_S

def getWizzardHelp():
  WizzardHelp= HelpSTXTopic_S('WizzardHelp', 'Wizzard Documentation',__doc__)
  return WizzardHelp


class Wizzard(Folder):
  '''A Wizzard manages a collection of forms used to provide step-by-step input for a complex task.
  For details, see the <a href="/Control_Panel/Products/Wizzard/Help/WizzardHelp">help</a>.'''
  meta_type= 'Wizzard'

  _properties= Folder._properties + (
    {'id': 'Dispatchers__', 'type': 'tokens', 'mode':'w',},
    )
  Dispatchers__= []

  manage_options= (
    ({'label' : 'Contents', 'action':'manage_main',
      'help' : ('Wizzard','WizzardHelp')},)
    + Folder.manage_options[1:]
    )

  _security= ClassSecurityInfo()

  _security.declareProtected('View',
                             'new', 'edit', 'clear', 'update',
                             'getValues', 'getCollection')


  def new(self, client=None, REQUEST=None, **kw):
    '''collect data for new object (most variables empty).'''
    return self._do('new', client, REQUEST, kw)
  newisDocTemp=1

  def edit(self, client=None, REQUEST=None, **kw):
    '''edit existing information (most variables taken from environment).'''
    return self._do('edit', client, REQUEST, kw)
  editisDocTemp=1

  def clear(self, client=None, REQUEST=None, **kw):
    '''clear the state.'''
    return self._do('clear', client, REQUEST, kw)
  clearisDocTemp=1

  def update(self, client=None, REQUEST=None, **kw):
    '''update the remembered values.'''
    return self._do('update', client, REQUEST, kw)
  updateisDocTemp=1


  def getValues(self):
    '''return the collection of values as a mapping object.'''
    d= DictStack(); i= 0; n= len(self.Dispatchers__)
    while i < n:
      i= i+1
      di= self.restrictedTraverse(self.Dispatchers__[n-i])
      d.push(di.getValues())
    return d.__of__(self)

  getCollection= getValues


  def _do(self, action, client, REQUEST, kw):
    '''do *action* for each dispatcher, then render.'''
    if REQUEST is None: REQUEST= self.REQUEST
    kwc= kw.copy(); kwc['DispatcherDoNotRender__']= 1
    for d in self.Dispatchers__:
      apply(getattr(self.restrictedTraverse(d),action),(client,REQUEST), kwc)
    dest= self._getDestination(REQUEST,  kw)
    if not dest: return
    del kwc['DispatcherDoNotRender__']
    kwc['DispatcherURL__']= "%s/%s" % (self._getBase(REQUEST, kw),dest)
    return apply(self.restrictedTraverse(dest).show,(client,REQUEST), kwc)

  
  def _getDestination(self, REQUEST, kw):
    a= self._get('WizzardDestination__', REQUEST,kw)
    if a is None: a= self.Dispatchers__[0]
    elif not a: return None
    return a


  def _getBase(self, REQUEST, kw):
    a= self._get('WizzardURL__', REQUEST,kw)
    if a is None: a= REQUEST['URL1']
    return a


  def _get(self, key, REQUEST, kw):
    marker= self
    v= kw.get(key,marker)
    if v is marker:
      if not REQUEST.has_key(key): return None
      v= REQUEST[v]
    return v

InitializeClass(Wizzard)


def manage_addWizzard(self,id,title='',Dispatchers__=(),REQUEST=None):
  '''constructor'''
  self= self.aq_parent.aq_parent # skip Wizzard(Product) and ProductWizzard
  o= Wizzard()
  o.id= id
  o.title= title
  o.Dispatchers__= Dispatchers__
  self._setObject(id,o)
  if REQUEST is not None:
    o= self._getOb(id)
    REQUEST.RESPONSE.redirect(o.absolute_url() + '/manage_main')


manage_addWizzardForm= DTMLFile('www/WizzardAdd', globals())



class DictStack(Explicit):
  '''a stack of mapping objects exposed as a mapping.'''

  _security= ClassSecurityInfo()

  __allow_access_to_unprotected_subobjects__= 1

  _security.declareObjectProtected('Access contents information')
  _security.declareProtected('View',
                             '__getitem__', 'get', 'has_key',
                             'keys', 'values', 'items')

  def __init__(self):
    self._stack= []

  def push(self,m):
    self._stack.append(m)

  def pop(self):
    self._stack.pop()

  def __getitem__(self,key):
    marker= self._stack
    v= self.get(key,marker)
    if v is marker:
      raise KeyError, key
    return v

  def get(self,key,default=None):
    s= self._stack
    i=0; n= len(s)
    while i < n:
      i= i+1
      v= s[n-1].get(key,s)
      if v is not s: return v
    return default

  def has_key(self,key):
    s= self._stack
    return self.get(key,s) is not s

  def keys(self):
    return self.getDict(self).keys()

  def values(self):
    return self.getDict(self).values()

  def items(self):
    return self.getDict().items()

  def getDict(self):
    d= {}
    for m in self._stack: d.update(m)
    return d

InitializeClass(DictStack)
